from __future__ import annotations

import logging
import textwrap
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Sequence, Optional, Tuple

import jinja2
import pandas as pd

from metricflow.protocols.sql_client import SqlClient
from metricflow.dag.mf_dag import DagNode, MetricFlowDag, NodeId, DisplayedProperty
from metricflow.dag.id_generation import EXEC_NODE_READ_SQL_QUERY, EXEC_NODE_WRITE_TO_TABLE
from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.visitor import Visitable

logger = logging.getLogger(__name__)


class ExecutionPlanTask(DagNode, Visitable, ABC):
    """A node (aka task) in the DAG representation of the execution plan.

    In the DAG, a node's parents represent the tasks that need to be run before the node can run. Using the term task
    for these nodes as it seems more intuitive.
    """

    def __init__(self, task_id: NodeId, parent_nodes: List[ExecutionPlanTask]) -> None:
        """Constructor.

        Args:
            task_id: the ID for the node
            parent_nodes: the nodes that should be executed before this one.
        """
        self._parent_nodes = parent_nodes
        super().__init__(node_id=task_id)

    @property
    def parent_nodes(self) -> Sequence[ExecutionPlanTask]:
        """Return the nodes that should execute before this one."""
        return self._parent_nodes

    @abstractmethod
    def execute(self) -> TaskExecutionResult:
        """Execute the actions of this node."""

    @property
    def task_id(self) -> NodeId:
        """Alias for node ID since the nodes represent a task"""
        return self.node_id

    @property
    @abstractmethod
    def sql_query(self) -> Optional[SqlQuery]:
        """If this runs a SQL query, return the associated SQL."""
        pass


@dataclass(frozen=True)
class SqlQuery:
    """A SQL query that can be run along with bind parameters."""

    sql_query: str
    bind_parameters: SqlBindParameters


@dataclass(frozen=True)
class TaskExecutionError(Exception):
    """Error if a task fails."""

    error_str: str


@dataclass(frozen=True)
class TaskExecutionResult:
    """The results of running a task."""

    start_time: float
    end_time: float
    errors: Tuple[TaskExecutionError, ...] = ()

    # If the task was an SQL query, it's stored here
    sql: Optional[str] = None
    bind_params: Optional[SqlBindParameters] = None
    # If the task produces a dataframe as a result, it's stored here.
    df: Optional[pd.DataFrame] = None


class SelectSqlQueryToDataFrameTask(ExecutionPlanTask):
    """A task that runs a SELECT and puts that result into a dataframe."""

    def __init__(  # noqa: D
        self,
        sql_client: SqlClient,
        sql_query: str,
        execution_parameters: SqlBindParameters,
        parent_nodes: Optional[List[ExecutionPlanTask]] = None,
    ) -> None:

        self._sql_client = sql_client
        self._sql_query = sql_query
        self._execution_parameters = execution_parameters
        super().__init__(task_id=self.create_unique_id(), parent_nodes=parent_nodes or [])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return EXEC_NODE_READ_SQL_QUERY

    @property
    def description(self) -> str:  # noqa: D
        return "Run a query and write the results to a data frame"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [DisplayedProperty(key="sql_query", value=self._sql_query)]

    @property
    def execution_parameters(self) -> SqlBindParameters:  # noqa: D
        return self._execution_parameters

    def execute(self) -> TaskExecutionResult:  # noqa: D
        start_time = time.time()
        df = self._sql_client.query(stmt=self._sql_query, sql_bind_parameters=self.execution_parameters)
        end_time = time.time()
        return TaskExecutionResult(
            start_time=start_time, end_time=end_time, sql=self._sql_query, bind_params=self.execution_parameters, df=df
        )

    @property
    def sql_query(self) -> Optional[SqlQuery]:  # noqa: D
        return SqlQuery(
            sql_query=self._sql_query,
            bind_parameters=self._execution_parameters,
        )

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(sql_query='{self._sql_query}')"


class SelectSqlQueryToTableTask(ExecutionPlanTask):
    """A task that runs a SELECT and puts that result into a table."""

    def __init__(  # noqa: D
        self,
        sql_client: SqlClient,
        sql_query: str,
        execution_parameters: SqlBindParameters,
        output_table: SqlTable,
        parent_nodes: Optional[List[ExecutionPlanTask]] = None,
    ) -> None:
        self._sql_client = sql_client
        self._sql_query = sql_query
        self._output_table = output_table
        self._execution_parameters = execution_parameters
        super().__init__(task_id=self.create_unique_id(), parent_nodes=parent_nodes or [])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return EXEC_NODE_WRITE_TO_TABLE

    @property
    def description(self) -> str:  # noqa: D
        return "Run a query and write the results to a table"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [
            DisplayedProperty(key="sql_query", value=self._sql_query),
            DisplayedProperty(key="output_table", value=self._output_table),
            DisplayedProperty(key="execution_parameters", value=self._execution_parameters),
        ]

    def execute(self) -> TaskExecutionResult:  # noqa: D
        start_time = time.time()
        logger.info(f"Dropping table {self._output_table} in case it already exists")
        self._sql_client.drop_table(self._output_table)
        logger.info(f"Creating table {self._output_table} using a SELECT query")
        self._sql_client.create_table_as_select(
            self._output_table,
            select_query=self._sql_query,
            sql_bind_parameters=self._execution_parameters,
        )
        end_time = time.time()
        return TaskExecutionResult(start_time=start_time, end_time=end_time, sql=self._sql_query)

    @property
    def sql_query(self) -> Optional[SqlQuery]:  # noqa: D
        query_text = jinja2.Template(
            textwrap.dedent(
                """\
                CREATE TABLE {{ output_table }} AS (
                  {{ select_query | indent(2) }}
                )
                """
            ),
            undefined=jinja2.StrictUndefined,
        ).render(output_table=self._output_table.sql, select_query=self._sql_query)

        return SqlQuery(
            sql_query=query_text,
            bind_parameters=self._execution_parameters,
        )

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(sql_query='{self._sql_query}', output_table={self._output_table})"


class ExecutionPlan(MetricFlowDag[ExecutionPlanTask]):
    """A DAG where the nodes are tasks, and parents represent prerequisite tasks."""

    def __init__(self, plan_id: str, leaf_tasks: List[ExecutionPlanTask]) -> None:
        """Constructor.

        Args:
            plan_id: A string to uniquely identify this plan.
            leaf_tasks: The final set of tasks that will run, after task dependencies are finished.
        """
        super().__init__(dag_id=plan_id, sink_nodes=leaf_tasks)

    @property
    def tasks(self) -> Sequence[ExecutionPlanTask]:
        """Return all tasks in this plan."""
        if len(self.sink_nodes) == 0:
            return []

        def recursively_get_tasks(task: ExecutionPlanTask) -> List[ExecutionPlanTask]:
            tasks_to_return = []
            for parent_node in task.parent_nodes:
                tasks_to_return.extend(recursively_get_tasks(parent_node))
            tasks_to_return.append(task)

            return tasks_to_return

        assert len(self.sink_nodes) == 1
        return recursively_get_tasks(self.sink_nodes[0])



import copy
import pytest
import re
from typing import Callable

from metricflow.model.model_validator import ModelValidator
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.specs import IdentifierSpec
from metricflow.test.test_utils import find_data_source_with


@pytest.mark.skip("TODO: re-enforce after validations improvements")
def test_lonely_identifier_raises_issue(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    model = copy.deepcopy(simple_model__pre_transforms)
    lonely_identifier_name = "hi_im_lonely"

    func: Callable[[DataSource], bool] = lambda data_source: len(data_source.identifiers) > 0
    data_source_with_identifiers, _ = find_data_source_with(model, func)
    data_source_with_identifiers.identifiers[0].name = IdentifierSpec.from_name(lonely_identifier_name).element_name
    model_validator = ModelValidator()
    build = model_validator.validate_model(model)

    found_warning = False
    warning = (
        f"Identifier `{lonely_identifier_name}` only found in one data source `{data_source_with_identifiers.name}` "
        f"which means it will be unused in joins."
    )
    if build.issues is not None:
        for issue in build.issues:
            if re.search(warning, issue.message):
                found_warning = True

    assert found_warning



import json
from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.data_warehouse import (
    TableProperties,
    ColumnProperties,
    InferenceColumnType,
    DataWarehouseInferenceContextProvider,
)


class SnowflakeInferenceContextProvider(DataWarehouseInferenceContextProvider):
    """The snowflake implementation for a DataWarehouseInferenceContextProvider"""

    COUNT_DISTINCT_SUFFIX = "countdistinct"
    COUNT_NULL_SUFFIX = "countnull"
    MIN_SUFFIX = "min"
    MAX_SUFFIX = "max"

    def _column_type_from_show_columns_data_type(self, type_str: str) -> InferenceColumnType:
        """Get the correspondent InferenceColumnType from Snowflake's returned type string.

        See for reference: https://docs.snowflake.com/en/sql-reference/sql/show-columns.html
        """
        str_dict = {
            "FIXED": InferenceColumnType.INTEGER,
            "TEXT": InferenceColumnType.STRING,
            "REAL": InferenceColumnType.FLOAT,
            "BOOLEAN": InferenceColumnType.BOOLEAN,
            "DATE": InferenceColumnType.DATETIME,
            "TIMESTAMP_TZ": InferenceColumnType.DATETIME,
            "TIMESTAMP_LTZ": InferenceColumnType.DATETIME,
            "TIMESTAMP_NTZ": InferenceColumnType.DATETIME,
        }

        return str_dict.get(type_str, InferenceColumnType.UNKNOWN)

    def _get_select_list_for_column_name(self, name: str, count_nulls: bool) -> str:
        statements = [
            f"COUNT(DISTINCT '{name}') AS {name}_{SnowflakeInferenceContextProvider.COUNT_DISTINCT_SUFFIX}",
            f"MIN('{name}') AS {name}_{SnowflakeInferenceContextProvider.MIN_SUFFIX}",
            f"MAX('{name}') AS {name}_{SnowflakeInferenceContextProvider.MAX_SUFFIX}",
            (
                f"SUM(CASE WHEN '{name}' IS NULL THEN 1 ELSE 0 END) AS {name}_{SnowflakeInferenceContextProvider.COUNT_NULL_SUFFIX}"
                if count_nulls
                else f"0 AS {name}_{SnowflakeInferenceContextProvider.COUNT_NULL_SUFFIX}"
            ),
        ]

        return ", ".join(statements)

    def _get_table_properties(self, table: SqlTable) -> TableProperties:
        all_columns_query = f"SHOW COLUMNS IN TABLE {table.sql}"
        all_columns = self._client.query(all_columns_query)

        sql_column_list = []
        col_types = {}
        col_nullable = {}
        select_lists = []

        for row in all_columns.itertuples():
            column = SqlColumn.from_names(
                db_name=row.database_name.lower(),
                schema_name=row.schema_name.lower(),
                table_name=row.table_name.lower(),
                column_name=row.column_name.lower(),
            )
            sql_column_list.append(column)

            type_dict = json.loads(row.data_type)
            col_types[column] = self._column_type_from_show_columns_data_type(type_dict["type"])
            col_nullable[column] = type_dict["nullable"]
            select_lists.append(
                self._get_select_list_for_column_name(
                    name=column.column_name,
                    count_nulls=col_nullable[column],
                )
            )

        select_lists.append("COUNT(*) AS rowcount")
        select_list = ", ".join(select_lists)
        statistics_query = f"SELECT {select_list} FROM {table.sql} SAMPLE ({self.max_sample_size} ROWS)"
        statistics_df = self._client.query(statistics_query)

        column_props = [
            ColumnProperties(
                column=column,
                type=col_types[column],
                is_nullable=col_nullable[column],
                null_count=statistics_df[f"{column.column_name}_{SnowflakeInferenceContextProvider.COUNT_NULL_SUFFIX}"][
                    0
                ],
                row_count=statistics_df["rowcount"][0],
                distinct_row_count=statistics_df[
                    f"{column.column_name}_{SnowflakeInferenceContextProvider.COUNT_DISTINCT_SUFFIX}"
                ][0],
                min_value=statistics_df[f"{column.column_name}_{SnowflakeInferenceContextProvider.MIN_SUFFIX}"][0],
                max_value=statistics_df[f"{column.column_name}_{SnowflakeInferenceContextProvider.MAX_SUFFIX}"][0],
            )
            for column in sql_column_list
        ]

        return TableProperties(table=table, column_props=column_props)



from __future__ import annotations

import re

from typing import Optional

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.base import HashableBaseModel


class Version(HashableBaseModel):  # noqa: D
    major: int
    minor: int

    _VERSION_REGEX = re.compile(r"^v[0-9]+\.[0-9]+$")

    @staticmethod
    def parse(version: str) -> Version:  # noqa: D

        if not Version._VERSION_REGEX.match(version):
            raise ParsingException(
                f"Expected a version of the form 'v<MAJOR_VERSION>.<MINOR_VERSION>' but got '{version}'."
            )
        if version[0] == "v":
            version = version[1:]

        parts = version.split(".")
        assert len(parts) == 2

        return Version(major=int(parts[0]), minor=int(parts[1]))

    def __str__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(major={self.major}, minor={self.minor})"


class SourceFile(HashableBaseModel):  # noqa: D
    path: str
    contents: str


class Commit(HashableBaseModel):  # noqa: D
    commit: str
    timestamp: int


class FileSlice(HashableBaseModel):  # noqa: D
    filename: str
    content: str
    start_line_number: int
    end_line_number: int


class YamlConfigFile(HashableBaseModel):
    """Serializable container for customer model YAML contents

    The serialization support is included here for scenarios where persisting the contents in non-filesystem storage
    services is necessary or desirable.
    """

    filepath: str
    contents: str
    url: Optional[str]


class SourceContext(HashableBaseModel):  # noqa: D
    definition_hash: str


class Metadata(HashableBaseModel):  # noqa: D
    repo_file_path: str
    file_slice: FileSlice



from abc import ABC, abstractmethod

from metricflow.model.objects.user_configured_model import UserConfiguredModel


class ModelTransformRule(ABC):
    """Encapsulates logic for transforming a model. e.g. add metrics based on measures."""

    @staticmethod
    @abstractmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:
        """Copy and transform the given model into a new model."""
        pass
